@@ -39,7 +39,14 @@
 #include <stdarg.h>
 #include <ctype.h>
 
+#ifdef HAVE_DEBCONF
 #include <cdebconf/debconfclient.h>
+#endif
+
+#ifdef HAVE_LIBSELINUX
+#include <selinux/selinux.h>
+#include <selinux/label.h>
+#endif
 
 #define DEFAULT_PASSWD_MASTER	"/usr/share/base-passwd/passwd.master"
 #define DEFAULT_GROUP_MASTER	"/usr/share/base-passwd/group.master"
@@ -78,7 +85,7 @@ const struct _info specialusers[] = {
     { 35, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* dos	*/
     { 36, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* msql	*/
     { 37, (FL_NOAUTOREMOVE)				},  /* operator */
-    { 41, (FL_KEEPHOME)					},  /* gnats	*/
+    { 41, (FL_KEEPHOME|FL_NOAUTOREMOVE)			},  /* gnats	*/
     { 70, (FL_NOAUTOREMOVE)				},  /* alias	*/
     { 71, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmaild	*/
     { 72, (FL_KEEPALL|FL_NOAUTOADD|FL_NOAUTOREMOVE)	},  /* qmails	*/
@@ -97,6 +104,7 @@ const struct _info specialgroups[] = {
     { 32, (FL_NOAUTOREMOVE)				},  /* postgres */
     { 35, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* dos	*/
     { 36, (FL_NOAUTOADD|FL_NOAUTOREMOVE)		},  /* msql	*/
+    { 41, (FL_NOAUTOREMOVE)				},  /* gnats	*/
     { 70, (FL_NOAUTOREMOVE)				},  /* qmail	*/
     { 0, 0}
 };
@@ -143,8 +151,98 @@ int		flag_debconf	= 0;
 const char*	user_domain	= DEFAULT_DEBCONF_DOMAIN;
 const char*	group_domain	= DEFAULT_DEBCONF_DOMAIN;
 
+#ifdef HAVE_DEBCONF
 struct debconfclient*	debconf	= NULL;
+#endif
 
+#ifdef HAVE_LIBSELINUX
+/* SELinux support */
+
+static struct selabel_handle* selabel_handle;
+
+static int selinux_setup(void) {
+
+    if (!is_selinux_enabled()) {
+        return 0;
+    }
+
+    /* Because this process is short lived, don't bother with selabel_close
+     * it isn't worth the additional complexity
+     */
+    selabel_handle = selabel_open(SELABEL_CTX_FILE, NULL, 0);
+    if (selabel_handle == NULL) {
+        const char * const reason = strerror(errno);
+
+        if (security_getenforce() != 0) {
+            fprintf(stderr, "SELinux: enabled but selabel_open failed: %s\n", reason);
+            return 1;
+        } else {
+            fprintf(stderr, "SELinux: enabled but in permissive mode, selabel_open failed: %s\n", reason);
+        }
+    }
+
+    return 0;
+}
+
+static int selinux_prepare_create_file(const char *path) {
+    char * ctx = NULL;
+    int    r   = 0;
+
+    if (selabel_handle == NULL) {
+        return 0;
+    }
+
+    if (selabel_lookup_raw(selabel_handle, &ctx, path, S_IFREG) != 0) {
+        if (errno == ENOENT) {
+            return 0;
+        }
+
+        const char * const reason = strerror(errno);
+
+        if (security_getenforce() != 0) {
+            fprintf(stderr, "SELinux: unable to lookup label for %s: %s\n", path, reason);
+            return 1;
+        } else {
+            fprintf(stderr, "SELinux: unable to lookup label for %s: %s\n", path, reason);
+            fprintf(stderr, "SELinux: ignoring error as in permissive mode, but a relabel is recommended.\n");
+            return 0;
+        }
+    }
+
+    r = setfscreatecon_raw(ctx);
+
+    freecon(ctx);
+
+    return r;
+}
+
+static void selinux_after_create_file(void) {
+    if (selabel_handle != NULL) {
+        int saved_errno = errno;
+        setfscreatecon_raw(NULL);
+        errno = saved_errno;
+    }
+}
+#else
+/* No SELinux support, provide stubs */
+
+static int selinux_setup(void) {
+    return 0;
+}
+
+static int selinux_prepare_create_file(const char *path) {
+    (void)path;
+
+    return 0;
+}
+
+static void selinux_after_create_file(void) {
+    return;
+}
+
+#endif
+
+#ifdef HAVE_DEBCONF
 /* Abort the program if talking to debconf fails.  Use ret exactly once. */
 #define DEBCONF_CHECK(ret)					\
     do {							\
@@ -162,6 +260,10 @@ struct debconfclient*	debconf	= NULL;
     DEBCONF_CHECK(debconf_register(debconf, (template), (question)))
 #define DEBCONF_SUBST(question, var, value) \
     DEBCONF_CHECK(debconf_subst(debconf, (question), (var), (value)))
+#else
+#define DEBCONF_REGISTER(template, question)
+#define DEBCONF_SUBST(question, var, value)
+#endif
 
 
 /* malloc() with out-of-memory checking.
@@ -621,6 +723,7 @@ void version() {
  * flag.  Aborts the problem on any failure.
  */
 int ask_debconf(const char* priority, const char* question) {
+#ifdef HAVE_DEBCONF
     int		ret;
     const char*	response;
 
@@ -640,6 +743,9 @@ int ask_debconf(const char* priority, const char* question) {
 	return 1;
     else
 	return 0;
+#else
+	return 0;
+#endif
 }
 
 
@@ -806,13 +912,13 @@ void process_old_entries(const struct _info* lst, struct _node** passwd, struct
 		free(id);
 	    }
 
+	    walk=walk->next;
 	    if (make_change) {
 		if (opt_verbose)
 		    printf("Removing %s \"%s\" (%u)\n", descr, oldnode->name, oldnode->id);
 		remove_node(passwd, oldnode);
 		flag_dirty++;
 	    }
-	    walk=walk->next;
 	    continue;
 	}
 	walk=walk->next;
@@ -836,7 +942,7 @@ void process_changed_accounts(struct _node* passwd, struct _node* group, struct
 	    continue;
 
 	mc=find_by_named_entry(master, passwd);
-	if (mc==NULL) 
+	if (mc==NULL)
 	    continue;
 
 	if (passwd->id!=mc->id) {
@@ -1039,7 +1145,15 @@ int write_passwd(const struct _node* passwd, const char* file) {
     if (opt_verbose>2)
 	printf("Writing passwd-file to %s\n", file);
 
-    if ((output=fopen(file, "wt"))==NULL) {
+    if (selinux_prepare_create_file(sys_passwd) != 0) {
+        return 0;
+    }
+
+    output = fopen(file, "wt");
+
+    selinux_after_create_file();
+
+    if (output == NULL) {
 	fprintf(stderr, "Failed to open passwd-file %s for writing: %s\n",
 		file, strerror(errno));
 	return 0;
@@ -1068,7 +1182,15 @@ int write_shadow(const struct _node* shadow, const char* file) {
     if (opt_verbose>2)
 	printf("Writing shadow-file to %s\n", file);
 
-    if ((output=fopen(file, "wt"))==NULL) {
+    if (selinux_prepare_create_file(sys_shadow) != 0) {
+        return 0;
+    }
+
+    output = fopen(file, "wt");
+
+    selinux_after_create_file();
+
+    if (output == NULL) {
 	fprintf(stderr, "Failed to open shadow-file %s for writing: %s\n",
 	       	file, strerror(errno));
 	return 0;
@@ -1110,7 +1232,15 @@ int write_group(const struct _node* group, const char* file) {
     if (opt_verbose>2)
 	printf("Writing group-file to %s\n", file);
 
-    if ((output=fopen(file, "wt"))==NULL) {
+    if (selinux_prepare_create_file(sys_group) != 0) {
+        return 0;
+    }
+
+    output = fopen(file, "wt");
+
+    selinux_after_create_file();
+
+    if (output == NULL) {
 	fprintf(stderr, "Failed to open group-file %s for writing: %s\n",
 		file, strerror(errno));
 	return 0;
@@ -1427,6 +1557,7 @@ int main(int argc, char** argv) {
     /* If DEBIAN_HAS_FRONTEND is set in the environment, we're running under
      * debconf.  Enable debconf prompting unless --dry-run was also given.
      */
+#ifdef HAVE_DEBCONF
     if (getenv("DEBIAN_HAS_FRONTEND")!=NULL && !opt_dryrun) {
 	debconf=debconfclient_new();
 	if (debconf==NULL) {
@@ -1435,6 +1566,7 @@ int main(int argc, char** argv) {
 	}
 	flag_debconf=1;
     }
+#endif
 
     if (read_passwd(&master_accounts, master_passwd)!=0)
 	return 2;
@@ -1465,23 +1597,28 @@ int main(int argc, char** argv) {
     if (opt_sanity)
 	return 0;
 
+    if (selinux_setup() != 0)
+        return 3;
+
     if (!opt_nolock && !opt_dryrun)
 	if (!lock_files())
-	    return 3;
+	    return 4;
 
     umask(0077);
 
     if (!commit_files()) {
 	unlock_files();
-	return 4;
+	return 5;
     }
 
     if (!opt_nolock && !opt_dryrun)
 	if (!unlock_files())
-	    return 5;
+	    return 6;
 
+#ifdef HAVE_DEBCONF
     if (debconf!=NULL)
 	debconfclient_delete(debconf);
+#endif
 
     if (opt_dryrun)
 	return flag_dirty;
